import os
import logging
import pandas as pd
from datasets import Dataset, DatasetDict, load_dataset, load_from_disk
from typing import Dict, Any, Optional
import os
import datetime


class DatasetProcessor:
    def __init__(
        self,
        cache_path: str = "petharbor_cache/",
    ):
        self.cache_path = cache_path
        self.logger = logging.getLogger(__name__)
        self._last_input_format = None  # Initialize

    def validate_dataset(self, dataset, text_column) -> None:
        if text_column not in dataset.column_names:
            error_message = f"Text column '{text_column}' not found in dataset. Please specifiy 'text_column' column to the class."
            self.logger.error(error_message)
            raise ValueError(error_message)
        # drop missing rows
        clean_dataset = dataset.filter(lambda example: example[text_column] is not None)
        self.logger.info(
            f"Dataset contains {len(dataset)} rows. After removing missing rows, {len(clean_dataset)} rows remain."
        )
        return clean_dataset

    def load_dataset_file(self, file_path: str, split: str = "train") -> Dataset:
        """
        Loads a dataset from a file or directory, attempting to use HuggingFace's
        `load_dataset` for known file types and `load_from_disk` for directories
        or other formats.

        Args:
            file_path (str): Path to the dataset file or directory.
            split (str): The split of the HuggingFace dataset to load (e.g., 'train', 'test', 'eval').
                        Defaults to 'train'.

        Returns:
            A HuggingFace Dataset object for the specified split.

        Raises:
            ValueError: If the file format is unsupported or the dataset/split cannot be loaded.
        """
        _, file_extension = os.path.splitext(file_path)
        file_extension = file_extension.lower()
        self._last_input_format = file_extension  # Store the input format
        _LOAD_MAPPING = {
            ".csv": "csv",
            ".arrow": "arrow",
            ".json": "json",
            ".parquet": "parquet",
        }
        if file_extension in _LOAD_MAPPING:
            dataset_format = _LOAD_MAPPING[file_extension]
            try:
                loaded_dataset = load_dataset(dataset_format, data_files=file_path)
                self.logger.info(f"Loaded '{dataset_format}' dataset from {file_path}")
            except Exception as e:
                self.logger.error(
                    f"Failed to load '{dataset_format}' dataset from {file_path}: {e}"
                )
                raise
        elif file_extension == ".pkl":
            try:
                loaded_dataset = pd.read_pickle(file_path)
                loaded_dataset = Dataset.from_pandas(loaded_dataset)
                loaded_dataset = DatasetDict({"train": loaded_dataset})
                self.logger.info(f"Loaded dataset from pickle file at {file_path}")
            except Exception as e:
                self.logger.error(
                    f"Failed to load dataset from pickle file at {file_path}: {e}"
                )
                raise
        else:
            try:
                loaded_dataset = load_from_disk(file_path)
                self.logger.info(f"Loaded dataset from disk at {file_path}")
            except Exception as e:
                self.logger.error(
                    f"Failed to load dataset from disk at {file_path}: {e}"
                )
                raise
        if not loaded_dataset:
            error_message = f"Dataset not found or could not be loaded from {file_path}"
            self.logger.error(error_message)
            raise ValueError(error_message)
        if split in loaded_dataset:
            self.logger.info(f"Returning '{split}' split from the loaded dataset.")
            return loaded_dataset[split]
        elif "train" in loaded_dataset and split != "train":
            self.logger.warning(
                f"Split '{split}' not found. Returning the default 'train' split. "
                f"Available splits are: {list(loaded_dataset.keys())}"
            )
            return loaded_dataset["train"]
        elif len(loaded_dataset.keys()) == 1:
            default_split = list(loaded_dataset.keys())[0]
            if split != default_split:
                self.logger.info(
                    f"Returning the only available split: '{default_split}'."
                )
            return loaded_dataset[default_split]
        else:
            error_message = (
                f"Split '{split}' not found in the loaded dataset. "
                f"Available splits are: {list(loaded_dataset.keys())}"
            )
            self.logger.error(error_message)
            raise ValueError(error_message)

    def load_cache(self, dataset, cache=False) -> tuple:
        """
        Filter out anonymised data from the dataset using a cache.

        Args:
            dataset: The dataset to filter.
            cache (bool | str): If True, removes examples where "annonymised" == 1.
                                If str, treats it as a column name and filters out rows
                                based on cached record IDs from a text file.
            cache_path (str): Path to cache directory (only used if `cache` is a str).

        Returns:
            tuple: (filtered_dataset, original_dataset)
        """
        target_dataset = dataset
        completed_dataset = None
        if cache:
            try:
                if isinstance(cache, bool):
                    target_dataset = dataset.filter(
                        lambda example: example.get("annonymised", 0) == 0
                    )

                    completed_dataset = dataset.filter(
                        lambda example: example.get("annonymised", 0) == 1
                    )

                elif isinstance(cache, str):
                    if cache not in dataset.column_names:
                        self.logger.error(f"Column '{cache}' not found in dataset.")
                        raise ValueError(f"Column '{cache}' not found in dataset.")

                    self.full_cache_path = os.path.join(
                        self.cache_path, f"{cache}_cache.txt"
                    )
                    if os.path.exists(self.full_cache_path):
                        with open(self.full_cache_path, "r") as f:
                            cached_ids = set(line.strip() for line in f)
                        target_dataset = dataset.filter(
                            lambda example: str(example[cache]) not in cached_ids
                        )

                        completed_dataset = dataset.filter(
                            lambda example: str(example[cache]) in cached_ids
                        )
                    else:
                        # Create the cache file if it doesn't exist
                        # create the directory if it doesn't exist
                        os.makedirs(os.path.dirname(self.cache_path), exist_ok=True)
                        self.cache_path = os.path.join(
                            self.cache_path, f"{cache}_cache.txt"
                        )
                        with open(self.full_cache_path, "w") as f:
                            f.write("")

                        self.logger.info(
                            f"Cache file '{self.cache_path}' created. No cached IDs found."
                        )
                        target_dataset = dataset
                        completed_dataset = None

                else:
                    raise ValueError(
                        "`cache` must be either a boolean or a string (column name)."
                    )

                self.logger.info(
                    f"Cache enabled | Skipping {len(completed_dataset)} Coded rows | Processing {len(target_dataset)} rows"
                )

            except Exception as e:
                self.logger.error(f"Failed to apply cache filtering: {e}")

            if not target_dataset:
                self.logger.info("All data appears to have been Coded. Exiting...")
                self.logger.warning(
                    "If this was unexpected, please check your cache file or delete a column called 'annonymised' in your dataset."
                )
                import sys

                sys.exit(0)
            else:
                self.logger.info(f"Processing {len(target_dataset)} non-Coded rows")
                return (
                    target_dataset,
                    completed_dataset,
                )
        else:
            self.logger.info("Cache disabled | Processing all data")
            return (
                target_dataset,
                completed_dataset,
            )

    def save_dataset_file(
        self,
        target_dataset: Dataset,
        completed_dataset: Dataset,
        output_dir: Optional[str] = None,
        cache: bool = False,
    ):
        """
        Save the target dataset to a file, attempting to use the same format
        as the original input.

        Args:
            original_data (Dataset): The original HuggingFace Dataset (used to infer format).
            target_dataset (Dataset): The HuggingFace Dataset to save.
            output_dir (str, optional): The directory or full file path for saving.
                                        If None, defaults to a CSV in the current directory.
            cache (bool or str, optional): If True, adds an 'annonymised' column and logs.
                                          If a string, appends the values from that column
                                          to the cache file. Defaults to False.
        """
        target_df = target_dataset.to_pandas()

        date = datetime.datetime.now().strftime("%Y_%m_%d")
        save_path = output_dir
        if cache:
            if isinstance(cache, bool):
                target_df["annonymised"] = 1
                self.logger.info(
                    "Cache enabled || 'annoymised' column added to dataset. Note: Review our documentation for more details."
                )
            elif isinstance(cache, str):
                cache_ids = target_df[cache].tolist()
                # Read in the cache file and append the new ids to the bottom
                with open(self.full_cache_path, "a") as f:
                    for id in cache_ids:
                        f.write(f"{id}\n")
            if completed_dataset is not None:
                original_df = completed_dataset.to_pandas()
                target_df = pd.concat([original_df, target_df], ignore_index=True)

            save_extensions = [".csv", ".json", ".parquet", ".arrow", ".pkl"]

            # Determine save path and format
            if output_dir and output_dir.endswith(tuple(save_extensions)):
                save_format = os.path.splitext(output_dir)[1]
                save_path = output_dir
                logging.info(f"Output format set to: {save_format}")
            elif output_dir:
                # If output_dir has no extension, assume it's a directory
                if hasattr(self, "_last_input_format") and self._last_input_format:
                    save_format = self._last_input_format
                    logging.info(
                        f"Using original input format for saving: {save_format}"
                    )
                else:
                    save_format = ".csv"
                    logging.info(
                        "No original input format found, defaulting to CSV for saving."
                    )
                base_name = f"petharbor_Coded_{date}{save_format}"
                save_path = os.path.join(output_dir, base_name)
            else:
                # Default if no output_dir provided
                save_format = ".csv"
                save_path = f"petharbor_Coded_{date}.csv"

            logging.info(f"Saving dataset using format: {save_format}")

            # Ensure output directory exists
            output_dirname = os.path.dirname(save_path)
            if output_dirname and not os.path.exists(output_dirname):
                os.makedirs(output_dirname, exist_ok=True)

            # Save based on format
            try:
                if save_format == ".csv":
                    target_df.to_csv(save_path, index=False)
                elif save_format == ".json":
                    target_df.to_json(save_path, orient="records", lines=True)
                elif save_format == ".parquet":
                    target_df.to_parquet(save_path, index=False)
                elif save_format == ".arrow":
                    target_dataset.save_to_disk(save_path)
                elif save_format == ".pkl":
                    target_df.to_pickle(save_path)
                else:
                    target_df.to_csv(save_path, index=False)
                    self.logger.warning(
                        f"Unsupported format '{save_format}', saved as CSV."
                    )

                self.logger.info(f"Saved dataset to {save_path} ({save_format}).")

            except Exception as e:
                self.logger.error(
                    f"Failed to save dataset to {save_path}: {e}. Aborting, saving output as petharbor_anonymised.csv to current directory"
                )
                target_df.to_csv("petharbor_anonymised.csv", index=False)
                raise



from pettag.utils.dataset import DatasetProcessor
from pettag.utils.processor_model import ModelProcessor
from pettag.utils.logging_setup import get_logger
from datasets import Dataset, load_dataset
from sentence_transformers import SentenceTransformer, util
from transformers import pipeline
from collections import defaultdict

from torch.nn import functional as F
from typing import Optional, Dict, Any
import torch
import logging
import pandas as pd
import numpy as np
import os
import json


class DiseaseCoder:
    """Anonymises text data using a pre-trained model.

    Args:
        dataset (Union[str, Dataset], optional): Path to dataset file (CSV, Arrow, etc.) or a HuggingFace Dataset.
        split (str): Dataset split to use ('train', 'test', etc.). Defaults to 'train'.
        model (str): HuggingFace model path or name. Defaults to 'SAVSNET/PetHarbor'.
        tokenizer (str, optional): Tokenizer path or name. Defaults to model if not provided.
        text_column (str): Column in dataset containing text. Defaults to 'text'.
        cache (bool): Whether to use caching. Defaults to True.
        cache_path (str): Directory to store cache files. Defaults to 'petharbor_cache/'.
        logs (str, optional): Path to save logs. If None, logs to console.
        device (str, optional): Device to use for computation. Defaults to 'cuda' if available else 'cpu'.
        tag_map (Dict[str, str], optional): Entity tag to replacement string mapping.
        output_dir (str, optional): Directory to save output dataset.
    """

    def __init__(
        self,
        framework: str = 'icd11', #options: 'icd11', 'icd10', 'snomed'
        dataset: str = None,  # Path to the dataset file (CSV, Arrow, etc.)
        split: str = "train",  # Split of the dataset to use (e.g., 'train', 'test', 'eval')
        model: str = "seanfarrell/bert-base-uncased",  # Path to the model
        tokenizer: str = None,  # Path to the tokenizer
        synonyms_dataset="seanfarrell/ICD-11_synonyms",
        synonyms_embeddings_dataset="cache/ICD-11_synonyms_embeddings.pt",
        embedding_model: str = "sentence-transformers/embeddinggemma-300m-medical",
        text_column: str = "text",  # Column name in the dataset containing text data
        label_column: str = "labels",  # Column name in the dataset containing labels
        cache: bool = True,  # Whether to use cache
        cache_path: str = "petharbor_cache/",  # Path to save cache files
        logs: Optional[str] = None,  # Path to save logs
        device: Optional[str] = "cuda:0" if torch.cuda.is_available() else "cpu",
        output_dir: str = None,  # Directory to save the output files
    ):
        self.framework = framework
        self.dataset = dataset
        self.split = split
        self.tokenizer = tokenizer if tokenizer else model
        self.text_column = text_column
        self.label_column = label_column
        self.cache = cache
        self.cache_path = cache_path
        self.logs = logs
        self.device = device
        self.output_dir = output_dir
        self.logger = self._setup_logger()

        self.dataset_processor = DatasetProcessor(cache_path=self.cache_path)
        self.logger.info(f"Initializing NER pipeline. Using {device}.")
        self.model = pipeline(
            "token-classification",
            model=model,
            tokenizer=self.tokenizer,
            aggregation_strategy="simple",
            device=self.device,
            dtype=torch.float16 if "cuda" in self.device else torch.float32,
        )
        self.logger.info("Initializing embedding pipeline")
        self.embedding_model = SentenceTransformer(embedding_model).to(self.device)

        try:
            self.logger.info(
                f"Loading precomputed ICD lookup from {synonyms_embeddings_dataset} ..."
            )
            data = torch.load(
                synonyms_embeddings_dataset, map_location=self.device, weights_only=True
            )
        except:
            self.logger.info(
                "No ICD embedding store found. Generating a new one. This should only happen on first run. Should take a few minutes..."
            )
            icd_lookup_daaset = load_dataset(synonyms_dataset, split="train", download_mode='force_redownload')
            data = self._preprocess_icd_lookup(
                disease_code_lookup=icd_lookup_daaset,
                save_path=synonyms_embeddings_dataset,
            )

        self.logger.info(f"Initializing {self.framework.upper()} ModelProcessor")
        self.model_processor = ModelProcessor(
            framework=self.framework,
            model=self.model,
            icd_embedding=data,
            text_column=self.text_column,
            label_column=self.label_column,
            embedding_model=self.embedding_model,
            device=self.device,
        )
        self.num_runs = 0

    def _setup_logger(self) -> Any:
        return get_logger(log_dir=self.logs) if self.logs else get_logger()

    def _print_output(self, input_text: str, output_text: str):
        timestamp = pd.Timestamp.now().strftime("%Y-%m-%d %H:%M:%S")
        pretty_output = json.dumps(output_text, indent=4, ensure_ascii=False, sort_keys=True)

        print(f"[{timestamp} | SUCCESS | PetCoder] Input:  {input_text}")
        print(f"[{timestamp} | SUCCESS | PetCoder] Output:\n{pretty_output}")

    def _prepare_single_text(self, text: str) -> Dataset:
        if not isinstance(text, str):
            error_message = "Input text must be a string."
            self.logger.error(error_message)
            raise ValueError(error_message)
        clean_text = text.strip()
        df = pd.DataFrame({self.text_column: [clean_text]})
        return Dataset.from_pandas(df)

        # -----------------------------------------------------

    # Create Embedding Store
    # -----------------------------------------------------

    def _preprocess_icd_lookup(self, disease_code_lookup, save_path="icd_lookup.pt"):
        """
        Preprocess an ICD lookup table into a compact, normalized, and efficiently loadable
        PyTorch dictionary. The resulting .pt file includes:
        - Normalized embeddings
        - Metadata (ICD11, ICD10, SNOMED)
        - Parent–subcode index mappings
        - .Z code mask

        Parameters
        ----------
        disease_code_lookup : pandas.DataFrame or dict-like
            ICD lookup data containing at least the columns:
            ['icd11Code', 'icd11Title', 'Title_synonym', 'icd11URI', 'embeddings',
            'ChapterNo', 'icd10Code', 'icd10Title', 'snomedCode', 'snomedTitle'].
        save_path : str, optional
            Output path for the saved lookup file (default: "icd_lookup.pt").

        Returns
        -------
        dict
            A dictionary containing tensors and metadata ready for downstream use.
        """

        self.logger.info("📦 Preprocessing ICD lookup into safe PyTorch format...")

        # -------------------------------------------------------------------------
        # ✅ Step 1. Normalize embeddings
        # -------------------------------------------------------------------------
        embeddings = torch.tensor( np.asarray(disease_code_lookup["embeddings"], dtype=np.float32) ) 
        embeddings = F.normalize(embeddings, dim=1)
        self.logger.info(f"Embeddings shape: {embeddings.shape}")

        # -------------------------------------------------------------------------
        # ✅ Step 2. Extract metadata columns
        # -------------------------------------------------------------------------
        icd11_codes = [str(c) for c in disease_code_lookup["icd11Code"]]
        icd11_titles = [str(t) for t in disease_code_lookup["icd11Title"]]
        synonyms = [str(s) for s in disease_code_lookup["Title_synonym"]]
        uris = [str(u) for u in disease_code_lookup["icd11URI"]]
        chapters = [str(ch) for ch in disease_code_lookup["ChapterNo"]]
        icd10_codes = [str(c) for c in disease_code_lookup["icd10Code"]]
        icd10_titles = [str(t) for t in disease_code_lookup["icd10Title"]]
        snomed_codes = [str(c) for c in disease_code_lookup["snomedCode"]]
        snomed_titles = [str(t) for t in disease_code_lookup["snomedTitle"]]

        # -------------------------------------------------------------------------
        # ✅ Step 3. Build parent → subcode mapping (based on ICD-11 code)
        # -------------------------------------------------------------------------
        parent_groups = {}
        for idx, code in enumerate(icd11_codes):
            parent = code.split(".")[0]
            parent_groups.setdefault(parent, []).append(idx)

        parent_to_subcodes_keys, parent_to_subcodes_values = [], []
        for parent, indices in parent_groups.items():
            if len(indices) > 1:
                parent_to_subcodes_keys.append(parent)
                parent_to_subcodes_values.append(torch.tensor(indices, dtype=torch.long))

        # -------------------------------------------------------------------------
        # ✅ Step 4. Create a ".Z" code mask (terminal ICD-11 codes)
        # -------------------------------------------------------------------------
        z_code_mask = torch.tensor([c.endswith(".Z") for c in icd11_codes], dtype=torch.bool)

        # -------------------------------------------------------------------------
        # ✅ Step 5. Handle save path and ensure directory exists
        # -------------------------------------------------------------------------
        save_path = save_path if save_path.endswith(".pt") else f"{save_path}.pt"
        os.makedirs(os.path.dirname(save_path) or ".", exist_ok=True)

        # -------------------------------------------------------------------------
        # ✅ Step 6. Save lookup dictionary
        # -------------------------------------------------------------------------
        lookup_dict = {
            "lookup_embeddings": embeddings.cpu(),
            "icd11Code": icd11_codes,
            "icd11Title": icd11_titles,
            "Title_synonym": synonyms,
            "icd11URI": uris,
            "ChapterNo": chapters,
            "icd10Code": icd10_codes,
            "icd10Title": icd10_titles,
            "snomedCode": snomed_codes,
            "snomedTitle": snomed_titles,
            "parent_to_subcodes_keys": parent_to_subcodes_keys,
            "parent_to_subcodes_values": parent_to_subcodes_values,
            "z_code_mask": z_code_mask,
        }

        torch.save(lookup_dict, save_path)
        self.logger.info(f"💾 ICD lookup saved successfully → {save_path}")

        # -------------------------------------------------------------------------
        # ✅ Step 7. Return device-ready dictionary
        # -------------------------------------------------------------------------
        return {
            key: value.to(self.device) if isinstance(value, torch.Tensor) else value
            for key, value in lookup_dict.items()
        }

    def _prepare_dataset(self) -> Dataset:
        if isinstance(self.dataset, Dataset):
            original_data = self.dataset
        elif isinstance(self.dataset, str):
            original_data = self.dataset_processor.load_dataset_file(
                self.dataset, split=self.split
            )
        else:
            raise ValueError("`dataset` must be a filepath or a HuggingFace Dataset.")

        validated = self.dataset_processor.validate_dataset(
            dataset=original_data, text_column=self.text_column
        )
        completed_dataset, target_dataset = self.dataset_processor.load_cache(
            dataset=validated, cache=self.cache
        )
        return completed_dataset, target_dataset

    def _run(
        self,
        text: str = None,
        dataset: str = None,
    ) -> None:
        """Coder the single text data or in a dataset and output/saves the results.

        Args:
        text (str, optional): Text to code
        dataset (str, optional): Path to the dataset file (CSV, Arrow, etc.)

        Raises:
        ValueError: If both text and dataset are provided or neither is provided.

        """
        self.num_runs += 1
        if text and self.dataset:
            raise ValueError(
                "Please provide either a text string or a dataset path, not both."
            )
        # Prepare input
        if text:  # If text is provided
            self.logger.warning(
                "Coding single text input. For bulk processing, use a dataset."
            )
            target_dataset = self._prepare_single_text(text)
        elif dataset:  # If dataset is provided to class
            self.dataset = dataset
            target_dataset, completed_dataset = self._prepare_dataset()
        elif self.dataset:  # If dataset is initialized
            target_dataset, completed_dataset = self._prepare_dataset()
        else:
            raise ValueError("Please provide either a text string or a dataset path.")
        if text:
            target_dataset = self.model_processor.single_predict(dataset=target_dataset)
            self._print_output(text, target_dataset[0])
            return target_dataset[0]
        else:
            target_dataset = self.model_processor.predict(
                dataset=target_dataset, batch_size=16
            )
            self.dataset_processor.save_dataset_file(
                target_dataset=target_dataset,
                completed_dataset=completed_dataset,
                cache=self.cache,
                output_dir=self.output_dir,
            )

    def predict(self, text: str = None, dataset: str = None) -> Optional[str]:
        """
        Anonymises text data.

        n.b 'anonymise' method overwrites the text column if a dataset is provided.
        If only text is provided, the anonymised text is returned.
        """
        if self.num_runs > 1:
            self.logger.warning(
                "It appears you are sequentially running the model. Reccomended: Pass 'dataset' to the class."
            )

        if dataset is None:
            dataset = self.dataset
        if text is not None:
            return self._run(text=text, dataset=None)
        elif dataset is not None:
            self._run(text=None, dataset=dataset)
            return None  # Explicitly return None when operating on the dataset
        else:
            self.logger.warning("No text or dataset provided for anonymisation.")
            return None

if __name__ == "__main__":
    # Example usage
    coder = DiseaseCoder()

    text = "Cookie present with vomiting and diarrhea. Suspected gastroenteritis."
    coder.predict(text.upper())


from pettag.utils.logging_setup import get_logger
import torch
from tqdm.contrib.logging import logging_redirect_tqdm
from torch.nn import functional as F
import pandas as pd
import numpy as np
from datasets import Dataset


class ModelProcessor:
    def __init__(
        self,
        framework,
        model,
        icd_embedding="icd_lookup.pt",  # new default filename
        replaced=True,
        text_column="text",
        label_column="ICD_11_code",
        embedding_model=None,
        device=None,
    ):
        self.model = model
        self.replaced = replaced
        self.text_column = text_column
        self.label_column = label_column
        self.embedding_model = embedding_model
        self.device = (
            device if device else ("cuda:0" if torch.cuda.is_available() else "cpu")
        )
        self.framework = framework
        self.logger = get_logger()

        self.lookup_embeddings = icd_embedding["lookup_embeddings"].to(self.device)
        self.icd11_codes = icd_embedding["icd11Code"]
        self.icd11_titles = icd_embedding.get("icd11Title", None)
        self.title_synonyms = icd_embedding.get("Title_synonym", None)
        self.icd11_uris = icd_embedding.get("icd11URI", None)
        self.chapters = icd_embedding.get("ChapterNo", None)
        self.icd10_codes = icd_embedding.get("icd10Code", None)
        self.icd10_titles = icd_embedding.get("icd10Title", None)
        self.snomed_codes = icd_embedding.get("snomedCode", None)
        self.snomed_titles = icd_embedding.get("snomedTitle", None)

        self.num_codes = len(self.icd11_codes)

        self.parent_to_subcodes = {
            k: v.to(self.device)
            for k, v in zip(
                icd_embedding["parent_to_subcodes_keys"],
                icd_embedding["parent_to_subcodes_values"],
            )
        }

        self.z_code_mask = icd_embedding["z_code_mask"].to(self.device)
        self.logger.info(f"✅ Loaded {self.num_codes} ICD codes on {self.device}.")

    # -----------------------------------------------------
    # Internal similarity computation
    # -----------------------------------------------------
    def _disease_coder_batch(self, encoded, Z_BOOST):
        sims = F.linear(encoded, self.lookup_embeddings)  # cosine similarities
        top_scores, top_idx = sims.max(dim=1)

        final_idx = top_idx.clone()
        final_score = top_scores.clone()

        # Vectorized subcode refinement
        for parent_code, sub_idx in self.parent_to_subcodes.items():
            mask = torch.tensor(
                [self.icd11_codes[i].split(".")[0] == parent_code for i in top_idx],
                device=self.device,
            )
            if mask.any():
                sub_sims = sims[mask][:, sub_idx]
                sub_sims = sub_sims + (self.z_code_mask[sub_idx] * Z_BOOST)
                best_sub = sub_sims.argmax(dim=1)
                final_idx[mask] = sub_idx[best_sub]
                final_score[mask] = sub_sims.gather(1, best_sub.unsqueeze(1)).squeeze()

        return final_idx, final_score

    # -----------------------------------------------------
    # Public disease coding interface
    # -----------------------------------------------------
    @torch.inference_mode()
    def disease_coder_batch(self, diseases, Z_BOOST=0.06, framework=None):
        """
        Encode a batch of disease descriptions and return the most similar ICD/SNOMED entries.

        Parameters
        ----------
        diseases : list[str]
            A list of free-text disease descriptions to map to codes.
        Z_BOOST : float, optional
            A boost applied to ".Z" codes to slightly prioritize terminal classifications.

        Returns
        -------
        list[dict]
            A list of dictionaries containing framework-specific metadata and similarity scores.
        """
        if not diseases:
            return []
        framework = framework or self.framework
        framework = framework.lower()
        valid_frameworks = ["icd11", "icd10", "snomed"]
        if framework not in valid_frameworks:
            raise ValueError(
                f"Invalid framework '{framework}'. Must be one of {valid_frameworks}."
            )

        # -------------------------------------------------------------------------
        # ✅ Step 1. Encode input diseases using the embedding model
        # -------------------------------------------------------------------------
        encoded = self.embedding_model.encode(
            diseases,
            convert_to_tensor=True,
            device=self.device if torch.cuda.is_available() else None,
            batch_size=128,
            show_progress_bar=False,
        )
        encoded = F.normalize(encoded, dim=1)

        # -------------------------------------------------------------------------
        # ✅ Step 2. Perform vector similarity search
        # -------------------------------------------------------------------------
        final_idx, final_score = self._disease_coder_batch(encoded, Z_BOOST)

        # -------------------------------------------------------------------------
        # ✅ Step 3. Compile metadata for top matches according to framework
        # -------------------------------------------------------------------------
        results = []
        for i, disease in enumerate(diseases):
            idx = int(final_idx[i].item())
            score = min(float(final_score[i].item()), 1.0)

            if framework == "icd11":
                code = self.icd11_codes[idx]
                title = (
                    self.icd11_titles[idx] if hasattr(self, "icd11_titles") else None
                )
                chapter = self.chapters[idx] if hasattr(self, "chapters") else None
                uri = self.icd11_uris[idx] if hasattr(self, "icd11_uris") else None

            elif framework == "icd10":
                code = self.icd10_codes[idx] if hasattr(self, "icd10_codes") else None
                title = (
                    self.icd10_titles[idx] if hasattr(self, "icd10_titles") else None
                )
                chapter = self.chapters[idx] if hasattr(self, "chapters") else None
                uri = ""
                synonym = ""

            elif framework == "snomed":
                code = self.snomed_codes[idx] if hasattr(self, "snomed_codes") else None
                title = (
                    self.snomed_titles[idx] if hasattr(self, "snomed_titles") else None
                )
                chapter = ""
                uri = ""
                synonym = ""
                # if the first character of the code is a ( then add warning message that this is actually an ICD 11 code
            if code and str(code.strip()).startswith("("):
                self.logger.warning(
                    f"⚠️ The matched {framework} code '{code}' is an ICD-11 code."
                )

            results.append(
                {
                    "Input Disease": disease,
                    "Framework": framework.upper(),
                    "Code": code,
                    "Title": title,
                    "Chapter": chapter,
                    "URI": (
                        f"https://icd.who.int/browse/2025-01/mms/en#{uri}"
                        if uri
                        else None
                    ),
                    "Similarity": round(score, 4),
                }
            )

        return results

    def disease_coder(self, disease, Z_BOOST=0.06):
        return self.disease_coder_batch([disease], Z_BOOST)[0]

    # -----------------------------------------------------
    # NER and batch processing
    # -----------------------------------------------------
    def _process_batch(self, examples):
        texts = [t.lower() for t in examples[self.text_column]]
        ner_results = self.model(texts)
        all_diseases = []
        disease_spans = []
        batch_symptoms, batch_pathogens = [], []

        for doc in ner_results:
            diseases, symptoms, pathogens = set(), set(), set()
            for ent in doc:
                label = ent["entity_group"]
                word = ent["word"]
                if label == "DISEASE":
                    diseases.add(word)
                elif label == "SYMPTOM":
                    symptoms.add(word)
                elif label == "ETIOLOGY":
                    pathogens.add(word)

            start = len(all_diseases)
            all_diseases.extend(diseases)
            disease_spans.append((start, len(all_diseases)))
            batch_symptoms.append(list(symptoms))
            batch_pathogens.append(list(pathogens))

        if all_diseases:
            coded = self.disease_coder_batch(all_diseases)
            batch_diseases = [coded[s:e] for s, e in disease_spans]
        else:
            batch_diseases = [[] for _ in range(len(texts))]

        return {
            "disease_extraction": batch_diseases,
            "pathogen_extraction": batch_pathogens,
            "symptom_extraction": batch_symptoms,
            # self.label_column: batch_diseases,
        }

    # -----------------------------------------------------
    def predict(self, dataset, batch_size=64):
        date_time = pd.Timestamp.now().strftime("%Y-%m-%d %H:%M:%S")
        processed_dataset = dataset.map(
            self._process_batch,
            batched=True,
            batch_size=batch_size,
            desc=f"[{date_time} | INFO | PetCoder]",
            load_from_cache_file=False,
        )
        self.logger.info("Predictions obtained and text coded successfully.")
        return processed_dataset

    def single_predict(self, dataset):
        dataset = dataset.select([0])
        processed = self._process_batch(examples=dataset)
        return Dataset.from_dict(
            {
                self.text_column: dataset[self.text_column],
                "Code": processed["disease_extraction"],
                "pathogen_extraction": processed["pathogen_extraction"],
                "symptom_extraction": processed["symptom_extraction"],
                # self.label_column: processed[self.label_column],
            }
        )



import logging
import logging.handlers
import sys
from pathlib import Path


# Log level mapping
LOG_LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}

# Custom log format
LOG_FORMAT = "[%(asctime)s | %(levelname)s | %(custom_word)s] %(message)s"
DATE_FORMAT = "%Y-%m-%d %H:%M:%S"


class CustomFormatter(logging.Formatter):
    def __init__(self, fmt, datefmt, custom_word):
        super().__init__(fmt, datefmt)
        self.custom_word = custom_word

    def format(self, record):
        # Rename INFO to INFO1 (you can extend this with a dict if needed)
        if record.levelno == logging.INFO:
            record.levelname = "  INFO "
        elif record.levelno == logging.WARNING:
            record.levelname = "WARNING"
        elif record.levelno == logging.ERROR:
            record.levelname = " ERROR "
        # Add the custom word to the record
        record.custom_word = self.custom_word
        return super().format(record)


# Colored logging for console output
class ColoredFormatter(CustomFormatter):
    COLORS = {
        "DEBUG": "\033[94m",  # Blue
        "INFO": "\033[0m",
        "WARNING": "\033[93m",  # Yellow
        " ERROR ": "\033[91m",  # Red
        "CRITICAL": "\033[95m",  # Magenta
        "RESET": "\033[0m",
    }

    def format(self, record):
        log_msg = super().format(record)
        return f"{self.COLORS.get(record.levelname, '')}{log_msg}{self.COLORS['RESET']}"


CUSTOM_WORD = "PetCoder"

def get_logger(name="petharbor", level="debug", log_dir=None):
    """Configures and returns a logger with both console and file handlers."""
    if log_dir:
        try:
            LOG_DIR = Path(log_dir)
            LOG_DIR.mkdir(exist_ok=True)
            LOG_FILE = f"{LOG_DIR}/{name}.log"
        except Exception as e:
            raise Exception(f"Failed to create log directory: {e}")

    level = LOG_LEVELS.get(level.lower(), logging.DEBUG)

    logger = logging.getLogger(name)
    logger.setLevel(level)

    # Clear handlers to avoid duplication
    if logger.hasHandlers():
        logger.handlers.clear()

    # Console handler
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(ColoredFormatter(LOG_FORMAT, DATE_FORMAT, CUSTOM_WORD))
    logger.addHandler(console_handler)

    if log_dir:
        file_handler = logging.handlers.RotatingFileHandler(
            LOG_FILE, maxBytes=5_000_000, backupCount=5, encoding="utf-8"
        )
        file_handler.setFormatter(CustomFormatter(LOG_FORMAT, DATE_FORMAT, CUSTOM_WORD))
        logger.addHandler(file_handler)

    # Prevent logs from being passed to the root logger
    logger.propagate = False

    return logger


# Usage example
if __name__ == "__main__":
    log = get_logger("test_logger", "debug")
    log.debug("This is a debug message.")
    log.info("This is an info message.")
    log.warning("This is a warning message.")
    log.error("This is an error message.")
    log.critical("This is a critical message.")
